package NoRFC; # $Id: NoRFC.pm 135 2020-06-16 07:30:29Z minus $
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC - Management library of noRFC and drafts documents

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use NoRFC;

=head1 ABSTRACT

norfc - program which performs management of noRFC documents and drafts

=head1 DESCRIPTION

This module provides norfc functionality

=head2 again

Internal use only

=head2 is_valid

    $app->is_valid("RFC-R1001010") or die ("Incorrect document number");

Returns 0 and sets error if specified document number is not valid,
and returns 1 if is valid

    is_valid("RFC-R1001010") or die ("Incorrect document number");

Returns 0 (without error setting) if specified document number is not valid,
and returns 1 if is valid

=head2 m11r_checkdigit_compute

    my $m11r = m11r_checkdigit_compute("123456789"); # 5
    my $m11r = $app->m11r_checkdigit_compute("123456789"); # 5
    my $m11r = __PACKAGE__->m11r_checkdigit_compute("123456789"); # 5

Returns M11R checkdigit by specified digits-string

See also L<Algorithm::CheckDigits::M11_015> and check_okpo()

=head2 store

    my $store = $self->store();

Returns store object

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<Algorithm::CheckDigits>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/$VERSION/;
$VERSION = '1.00';

use base qw/CTK::App/;

use Carp;
use Encode;
use Encode::Locale;

use CTK::Util qw/ trim dtf tz_diff variant_stf /;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;

use YAML::XS qw/DumpFile LoadFile/;
use Try::Tiny;
use File::Spec;

use NoRFC::Store;

use constant {
        PROJECTNAME         => "NoRFC",
        DATADIR             => "norfc",
        YFILE               => "norfc.yml",
        DBFILE              => "norfc.db",
        RFCNUM_BASE_FORMAT  => "%d%03d%02d",
        RFCNUM_FORMAT       => "%s-%s%d%03d%02d%d",
        DEFAULT_CATEGORY    => "Unknown",
        DEFAULT_TITLE       => "No title",
        ROW_COUNT           => 1000,
    };

__PACKAGE__->register_handler(
    handler     => "version",
    description => sprintf("%s Version", PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    printf("%s/%s\n", $self->prefix, $self->VERSION);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "info",
    description => sprintf("%s info", PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    printf("%s %s information\n", $self->project, $self->VERSION);
    printf("Data dir        : %s\n", $self->datadir);
    printf("Status File     : %s\n", File::Spec->catfile($self->datadir, YFILE));
    printf("Config          : %s\n", $self->conf("loadstatus") ? $self->configfile : "not loaded");
    $self->debug(_explain($self->config)) if $self->conf("loadstatus");
    printf("DB status       : %s\n", $store->status ? "OK" : "ERROR");
    printf("DB error        : %s\n", $store->error || sprintf("Store (%s) unknown error", $store->dsn))
        unless $store->status;
    printf("DB DSN          : %s\n", $store->dsn);
    printf("Total documents : %s\n", $store->status ? $store->count : "unknown");

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "gen",
    description => "Generation noRFC or Draft document number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Files
    my $yfile = File::Spec->catfile($self->datadir, YFILE);
    my $dbfile = File::Spec->catfile($self->datadir, DBFILE);

    # Letter
    my $let = "";
    $let .= "R" if $self->option("regular");
    $let .= "P" if $self->option("public");
    $let .= "X" if $self->option("extra");
    $let .= "Z" if $self->option("private");
    $let ||= uc(trim($self->cli_prompt('Please type a char of the status (R/P/X/Z):','R')));
    if ($let !~ /^[RPXZ]$/) {
        $self->error("Incorrect the letter char selected! Please specify one flag -R, -P, -X or -Z only! Default: -R");
        return 0;
    }
    $self->debug(sprintf("Letter          : %s", $let));

    # Date data
    my ($mday, $mon, $year, $yday) = (localtime(time))[3,4,5,7];
    $mon++; $yday++; $year += 1900;

    # Year
    if (my $y = $self->option("year")) {
        if ($y =~ /^[0-9]{4}$/) {
            $year = $y;
        } else {
            $self->error("Incorrect year! Please specify real year");
            return 0;
        }
    }
    my $y3 = $year % 1000;
    $self->debug(sprintf("Year (Y3)       : %d (%d)", $year, $y3));
    $self->debug(sprintf("Month           : %d", $mon));

    # Year's day
    if (my $d = $self->option("day")) {
        if (is_int($d) && $d && $d <= 366) {
            $yday = $d;
        } else {
            $self->error("Incorrect year's day! Please specify real year's day (value between 1 and 365 or 366)");
            return 0;
        }
    }
    $self->debug(sprintf("Day (YDAY)      : %d (%d)", $mday, $yday));

    # Key
    my $dkey = sprintf "%d-%02d-%02d", $year, $mon, $mday;

    # Sequence number (counter)
    my $cnt = 0;
    if (my $number = $self->option("number")) {
        if (is_int($number) && $number && $number <= 99) {
            $cnt = $number;
        } else {
            $self->error("Incorrect number value! Please specify value between 1 and 99");
            return 0;
        }
    } else {
        # Load history data
        my $history = {};
        if (-e $yfile and -s _) {
            try {
                $history = LoadFile($yfile);
            } catch {
                $self->error(sprintf("Can't load file %s: %s", $yfile, $_ // "unknown error"));
                return 0;
            };
        }
        $cnt = ($history->{$dkey} || 0) + 1;
        if ($cnt > 99) {
            $self->error("No free counter values. Try specify year or day manually");
            return 0;
        }
    }

    # Document type
    my $type = "RFC";
    if (my $t = $self->option("type")) {
        if ($t =~ /draft/i) {
            $type = "Draft";
        } elsif ($t =~ /rfc/i) {
            $type = "RFC";
        } else {
            $self->error("Incorrect document type! Please specify value one from: Draft or RFC");
            return 0;
        }
    }
    $self->debug(sprintf("Document type   : %s", $type));

    # Category name
    my $_category = $self->option("category")
        || $self->cli_prompt('Please provide the category of the document:', DEFAULT_CATEGORY);
    my $category = decode(locale => $_category);
    if (length($category) > 64) {
        $self->error("The category value is too long! Maximum: 64 chars");
        return 0;
    }
    $self->debug(sprintf("Category        : %s", $_category));

    # Title
    my $_title = join(" ", @arguments)
        || $self->cli_prompt('Please provide the title of the document:', DEFAULT_TITLE);
    my $title = decode(locale => $_title);
    if (length($title) > 255) {
        $self->error("The title value is too long! Maximum: 255 chars");
        return 0;
    }
    $self->debug(sprintf("Title           : %s", $_title));

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Generate number
    my $result = "";
    while ($cnt <= 99) {
        $result = sprintf(RFCNUM_FORMAT, $type, $let, $y3, $yday, $cnt,
            m11r_checkdigit_compute(sprintf(RFCNUM_BASE_FORMAT, $y3, $yday, $cnt)));
        my %info = $store->get($result);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
            return 0;
        }
        if ($self->option("force")) {
            $self->debug(sprintf("Counter (force) : %d", $cnt));
            last;
        } elsif ($info{id}) { # Found
            $self->debug(sprintf("Counter skipped : %d", $cnt));
        } else {
            $self->debug(sprintf("Counter         : %d", $cnt));
            last;
        }
    } continue {
        $cnt++;
    }
    if ($cnt > 99) {
        $self->error("No free counter values. Try specify year or day manually");
        return 0;
    }
    $self->debug(sprintf("Document number : %s", $result));

    # Fix
    $store->add(
        num         => $result,
        pub         => time(),
        category    => $category,
        title       => $title,
    ) or do {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0 unless $self->option("force");
    };

    # Save history data
    try {
        DumpFile($yfile, {$dkey => $cnt});
    } catch {
        $self->error(sprintf("Can't write file %s: %s", $yfile, $_ // "unknown error"));
        return 0 unless $self->option("force");
    };

    # Show result
    print $result, "\n";

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "check",
    description => "Checking noRFC or Draft document number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):', "none"));
    $self->debug(sprintf("Document number : %s", $recipient));
    my $status = is_valid($recipient);

    # Get number digits
    my $digs;
    $digs = $1 if $recipient =~ /([0-9]+)/;
    unless ($digs && length($digs) > 1) {
        $self->error("The document number is incorrect");
        return 0;
    }

    # Parse number
    my ($test, $digit) = ($digs =~ /^(\d+?)(\d)$/);
    $self->debug(sprintf("Test string     : %s", $test));
    $self->debug(sprintf("Got digit       : %d", $digit));
    my $m11r = m11r_checkdigit_compute($test);
    $self->debug(sprintf("Expected digit  : %d", $m11r));

    # DB check
    if ($status) {
        my %info = $store->get($recipient);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
            return 0;
        }
        if ($info{id}) { # Found
            $self->debug(sprintf("Found document  : %d", $info{id}));
            $self->debug(sprintf("  Published     : %s",
                $info{pub} ? dtf("%YYYY-%MM-%DD %hh:%mm:%ss ".tz_diff(), $info{pub}) : "unknown"));
            $self->debug(sprintf("  Category      : %s", encode(locale => $info{category} || DEFAULT_CATEGORY)));
            $self->debug(sprintf("  Title         : %s", encode(locale => $info{title} || DEFAULT_TITLE)));
        }
    }

    # Status
    $self->debug(sprintf("Summary status  : %s", $status ? "PASSED" : "FAILED"));

    if ($digit == $m11r) {
        print "PASSED! Document number is valid\n";
    } else {
        print "FAILED! Document number is NOT valid\n";
        return 0;
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "show",
    description => "Search noRFC or Draft document by number or show list of documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Check document number
    my $recipient = trim($self->option("recipient") || shift(@arguments) || "all");
    $self->debug(sprintf("Document number : %s", $recipient)) if $recipient ne "all";
    if ($recipient ne "all") { # By num
        return 0 unless $self->is_valid($recipient);
        my %info = $store->get($recipient);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
            return 0;
        }
        printf("Document number : %s\n", $recipient);
        printf("  Id            : %s\n", $info{id} || 0);
        printf("  Published     : %s\n", $info{pub}
            ? dtf("%YYYY-%MM-%DD %hh:%mm:%ss ".tz_diff(), $info{pub})
            : "unknown");
        printf("  Category      : %s\n", encode(locale => $info{category} || DEFAULT_CATEGORY));
        printf("  Title         : %s\n", encode(locale => $info{title} || DEFAULT_TITLE));
        return 1;
    }

    # List!
    my $count = $store->count();
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }
    my $offset = $self->option("offset") || 0;
       $offset = 0 unless is_int($offset);
    my $row_count = $self->option("limit") || ROW_COUNT;
       $row_count = ROW_COUNT unless is_int($row_count);
    my $page = $self->option("page") || 0;
       $page = 0 unless is_int($page);
    my $pages = (($count - ($count % $row_count)) / $row_count) + (($count % $row_count) ? 1 : 0);

    # Test page
    if ($page && $page > $pages) {
        $self->error( sprintf("Incorrect page number specified. Range: 1-%d", $pages) );
        return 0;
    }
    if ($page && !$offset) {
        $offset = ($page - 1) * $row_count;
    } elsif (!$page && $offset) {
        $page = (( $offset - ($offset % $row_count)) / $row_count) + 1;
    }

    $self->debug(sprintf("Count (total)   : %d", $count));
    $self->debug(sprintf("Pages (total)   : %d", $pages));
    $self->debug(sprintf("Offset          : %s", $offset));
    $self->debug(sprintf("Row count       : %s", $row_count));
    $self->debug(sprintf("Page            : %d", $page || 1));

    # Get
    my @all = $store->getall($offset, $row_count);
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Show table
    my @arr;
    my $total = 0;
    my $totalf = "";
    print "------------------+---------------+-------------------------------------------\n";
format STDOUT_TOP =
 Number           | Category      | Title
------------------+---------------+-------------------------------------------
.
format STDOUT =
 @<<<<<<<<<<<<<<< | @<<<<<<<<<<<< | @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
@arr
.
format STDOUTBOT =
------------------+---------------+--------------------------------------------
@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
$totalf
.
    foreach my $row (@all) {
        $total++;
        @arr = ();
        push @arr, $row->[1] || '';
        push @arr, variant_stf(encode(locale => $row->[3] || ''), 13);
        push @arr, variant_stf(encode(locale => $row->[4] || ''), 42);
        write;
    }
    $~ = "STDOUTBOT";
    $totalf = sprintf("row(s) %d of %d; page %d of %d", $total, $count, $page || 1, $pages || 1);
    write;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "del",
    description => "Delete noRFC or Draft document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Check document number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):', "all"));
    $self->debug(sprintf("Document number : %s", $recipient));
    return 0 unless $self->is_valid($recipient);


    # delete
    my $store = $self->store;
    $store->del($recipient) or do {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    };

    # Show result
    printf "The document %s has been successfully removed\n", $recipient;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "truncate",
    description => "Delete all records from database",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    if ($self->cli_prompt(sprintf("Delete all records from database (%s)?:", $store->dsn), 'no') !~ /^\s*y/i) {
        $self->error("Operation aborted");
        return 0;
    }

    # Truncate
    $store->truncate() or do {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    };

    # Show result
    printf "Database was cleared\n";

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "set",
    description => "Add or replace the noRFC or Draft document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Check document number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):', "all"));
    #$self->debug(sprintf("Document number : %s", $recipient));
    return 0 unless $self->is_valid($recipient);

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    my %info = $store->get($recipient);
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Category name
    my $_category = $self->option("category")
        || $self->cli_prompt('Please provide the category of the document:', $info{category} || DEFAULT_CATEGORY);
    my $category = decode(locale => $_category);
    if (length($category) > 64) {
        $self->error("The category value is too long! Maximum: 64 chars");
        return 0;
    }
    $info{category} = $category;

    # Title
    my $_title = join(" ", @arguments)
        || $self->cli_prompt('Please provide the title of the document:', $info{title} || DEFAULT_TITLE);
    my $title = decode(locale => $_title);
    if (length($title) > 255) {
        $self->error("The title value is too long! Maximum: 255 chars");
        return 0;
    }
    $info{title} = $title;

    # Set new data by old data
    $info{num} ||= $recipient;
    $info{pub} = time();
    printf("Document number : %s\n", $recipient);
    printf("  Id            : %s\n", $info{id}) if $info{id};
    printf("  Published     : %s\n", dtf("%YYYY-%MM-%DD %hh:%mm:%ss ".tz_diff(), $info{pub}));
    printf("  Category      : %s\n", $_category);
    printf("  Title         : %s\n", $_title);
    if ($self->cli_prompt('All right?:','yes') !~ /^\s*y/i) {
        $self->error("Operation aborted");
        return 0;
    }

    # Go!
    my $st = $info{id} ? $store->set(%info) : $store->add(%info);
    unless ($st) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Show result
    printf "The document %s has been successfully saved\n", $recipient;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "export",
    description => "Export all noRFC or Draft documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Permissions
    if (!$self->option("force") && -t STDOUT) { # Is TTy or force mode
        $self->error("Please use STDOUT pipe mode or set force flag (-f) for export");
        return 0;
    }
    if ($self->debugmode) {
        $self->error("Debug and verbose mode do not supported here");
        return 0;
    }

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Get count
    my $count = $store->count;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }
    #$self->debug(sprintf("Count           : %d", $count));

    # Create csv object
    require Text::CSV;
    binmode(STDOUT, ":raw:utf8");
    my $csv = Text::CSV->new();
    #$csv->decode_utf8(1);

    # Get data
    my $step = ROW_COUNT;
    for (my $offset = 0; $offset < $count; $offset += $step) {
        my $row_count = ($offset + $step > $count) ? $count % $step : $step;
        #$self->debug( sprintf("  > Offset=%d; count=%d", $offset, $row_count) );

        # Get
        my @all = $store->getall($offset, $row_count);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
            return 0;
        }

        # Print
        $csv->say(*STDOUT, $_) for (@all);
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "import",
    description => "Import all noRFC or Draft document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Permissions
    if (!$self->option("force") && -t STDIN) { # Is TTy or force mode
        $self->error("Please use STDIN pipe mode or set force flag (-f) for export");
        return 0;
    }
    if ($self->debugmode) {
        $self->error("Debug and verbose mode do not supported here");
        return 0;
    }

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
        return 0;
    }

    # Create csv object
    require Text::CSV;
    binmode(STDIN, ":raw:utf8");
    my $csv = Text::CSV->new();

    my $all = 0;
    my $errs = 0;
    my $added = 0;
    my $updated = 0;
    while (my $row = $csv->getline (*STDIN)) {
        #print(_explain($row));
        my %info = $store->get($row->[1]);
        my %info_new = (
                num => $row->[1],
                pub => $row->[2] || time(),
                category => $row->[3] || DEFAULT_CATEGORY,
                title => $row->[4] || DEFAULT_TITLE,
            );
        my $st = 0;
        if ($info{id}) {
            $st = $store->set(%info_new);
            $updated++ if $st;
        } else {
            $st = $store->add(%info_new);
            $added++ if $st;
        }
        if ($st) {
            $all++;
        } else {
            $errs++;
            printf STDERR "%s> %s\n", $row->[1] || "none", $store->error || sprintf("Store (%s) unknown error", $store->dsn);
            next;
        }
    }

    # Show result
    printf "Imported documents\t: %d\n\tAdded\t\t: %d\n\tUpdated\t\t: %d\n\tErrors\t\t: %d\n", $all, $added, $updated, $errs;

    return 1;
});


sub again {
    my $self = shift;
       $self->SUPER::again(); # CTK::App again first!!

    # Reset datadir from config file
    if (my $newdatadir = $self->config("datadir")) {
        $self->datadir($newdatadir);
    }

    # Store
    my $store_conf = $self->config("store") || {};
       $store_conf = {} unless is_hash($store_conf);
    my %store_args = %$store_conf;
    $store_args{file} = File::Spec->catfile($self->datadir, DBFILE)
        unless ($store_args{file} || $store_args{dsn});
    my $store = new NoRFC::Store(%store_args);
    $self->{store} = $store;
    #$self->debug(_explain($store));

    return $self; # CTK requires!
}
sub store {
    my $self = shift;
    return $self->{store}
}
sub m11r_checkdigit_compute {
    # See also: Algorithm::CheckDigits::M11_015 and check_okpo()
    my $test = pop;
    croak "Incorrect input digit-string" if !$test || $test =~ m/[^0-9]/g;
    my $len = length($test);
    my $iters = ($len + (($len & 1) ? 1 : 0)) / 2;
    my @digits = split(//, $test); # Get all digits from input string of chars
    #printf "Test=%s; len=%d; iters=%d\n", $test, $len, $iters;

    my $w_lim = 10; # Maximum for round-robin(10) weight list: 1,2,3,4,5,6,7,8,9,10,1,2,3,4,5...
    my $step = 2; # Step for weight list offset calculation for next iteration

    # Calculation sum for one weight list by ofset
    my $calc = sub {
        my $off = shift || 0;
        my $s = 0;
        for (my $i = 0; $i < $len; $i++) {
            my $w = (($i + $off) % $w_lim) + 1;
            $s += ($w * $digits[$i]);
            #printf " > i=%d; d=%d; w=%d; sum=%d\n", $i, $digits[$i], $w, $s;
        }
        return $s % 11;
    };

    # Main cycle
    my $sum = 0;
    for (my $j = 0; $j < $iters; $j++) {
        my $offset = $j*$step;
        $sum = $calc->($offset);
        #printf " >> j=%d; offset=%d; sum=%d\n", $j, $offset, $sum;
        last if $sum < 10;
    }
    $sum = 0 if $sum >= 10; # 0 if incorrect again
    #printf " >>> sum=%d\n", $sum;

    return $sum;
}
sub is_valid {
    my $number = pop;
    my $self = shift;
    unless ($number) {
        $self->error("The document number is incorrect") if $self;
        return 0;
    }

    # Format valid
    my $test = 0;
    my $exp = 0;
    if ($number =~ /^(Draft|RFC)\-([RPXZ])([0-9]{1,3})([0-9]{3})([0-9]{2})([0-9]{1})$/) {
        unless ($3) { # Y3
            $self->error("The Y3 field of document number is incorrect") if $self;
            return 0;
        }
        unless ($4) { # YDAY
            $self->error("The YDAY field of document number is incorrect") if $self;
            return 0;
        }
        unless ($5) { # NN
            $self->error("The NN field of document number is incorrect") if $self;
            return 0;
        }
        $test = sprintf("%d%03d%02d", $3, $4, $5);
        $exp = $6; # Expected
    }

    # M11R valid
    my $m11r = m11r_checkdigit_compute($test);
    if ($exp != $m11r) {
        $self->error("Document number is NOT valid") if $self;
        return 0;
    }

    return 1; # OK
}

sub _explain {
    require Data::Dumper;
    my $dumper = Data::Dumper->new ( [shift] );
    $dumper->Indent(1)->Terse(1);
    $dumper->Sortkeys(1) if $dumper->can("Sortkeys");
    return $dumper->Dump;
}

1;

__END__
